from unittest import TestCase

from rope.contrib.autoimport import models


class QueryTest(TestCase):
    def test_select_non_existent_column(self):
        with self.assertRaisesRegex(ValueError, """Unknown column names passed: {['"]doesnotexist['"]}"""):
            models.Name.objects.select('doesnotexist')._query


class NameModelTest(TestCase):
    def test_name_objects(self):
        self.assertEqual(
            models.Name.objects.select_star()._query,
            "SELECT * FROM names",
        )

    def test_query_strings(self):
        with self.subTest("objects"):
            self.assertEqual(
                models.Name.objects.select_star()._query,
                'SELECT * FROM names',
            )

        with self.subTest("search_submodule_like"):
            self.assertEqual(
                models.Name.search_submodule_like.select_star()._query,
                'SELECT * FROM names WHERE module LIKE ("%." || ?)',
            )

        with self.subTest("search_module_like"):
            self.assertEqual(
                models.Name.search_module_like.select_star()._query,
                'SELECT * FROM names WHERE module LIKE (?)',
            )

        with self.subTest("import_assist"):
            self.assertEqual(
                models.Name.import_assist.select_star()._query,
                "SELECT * FROM names WHERE name LIKE (? || '%')",
            )

        with self.subTest("search_by_name_like"):
            self.assertEqual(
                models.Name.search_by_name_like.select_star()._query,
                'SELECT * FROM names WHERE name LIKE (?)',
            )

        with self.subTest("delete_by_module_name"):
            self.assertEqual(
                models.Name.delete_by_module_name._query,
                'DELETE FROM names WHERE module = ?',
            )


class PackageModelTest(TestCase):
    def test_query_strings(self):
        with self.subTest("objects"):
            self.assertEqual(
                models.Package.objects.select_star()._query,
                'SELECT * FROM packages',
            )

        with self.subTest("delete_by_package_name"):
            self.assertEqual(
                models.Package.delete_by_package_name._query,
                'DELETE FROM packages WHERE package = ?',
            )



import pathlib
import tempfile

import pytest

from rope.base import libutils, resources, pyobjectsdef
from rope.base.project import Project
from ropetest import testutils


@pytest.fixture
def project():
    proj = testutils.sample_project()
    yield proj
    testutils.remove_project(proj)


@pytest.fixture
def mod1(project):
    testutils.create_package(project, "pkg1")
    return testutils.create_module(project, "pkg1.mod1")


def test_repr_project():
    with tempfile.TemporaryDirectory() as folder:
        folder = pathlib.Path(folder).resolve()
        obj = testutils.sample_project(folder)
        assert isinstance(obj, Project)
        assert repr(obj) == f'<rope.base.project.Project "{folder}">'


def test_repr_file(project):
    obj = project.get_file("test/file.py")
    assert isinstance(obj, resources.File)
    assert repr(obj).startswith('<rope.base.resources.File "test/file.py" at 0x')


def test_repr_folder(project):
    obj = project.get_folder("test/folder")
    assert isinstance(obj, resources.Folder)
    assert repr(obj).startswith('<rope.base.resources.Folder "test/folder" at 0x')


def test_repr_pyobjectsdef_pymodule(project, mod1):
    obj = project.get_module("pkg1.mod1")
    assert isinstance(obj, pyobjectsdef.PyModule)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyModule "pkg1.mod1" at 0x')


def test_repr_pyobjectsdef_pymodule_without_associated_resource(project):
    obj = pyobjectsdef.PyModule(project.pycore, "a = 1")
    assert isinstance(obj, pyobjectsdef.PyModule)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyModule "" at 0x')


def test_repr_pyobjectsdef_pypackage(project, mod1):
    obj = project.get_module("pkg1")
    assert isinstance(obj, pyobjectsdef.PyPackage)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyPackage "pkg1" at 0x')


def test_repr_pyobjectsdef_pypackage_without_associated_resource(project, mod1):
    obj = pyobjectsdef.PyPackage(project.pycore)
    assert isinstance(obj, pyobjectsdef.PyPackage)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyPackage "" at 0x')


def test_repr_pyobjectsdef_pyfunction(project, mod1):
    code = """def func(arg): pass"""
    mod = libutils.get_string_module(project, code, mod1)
    obj = mod.get_attribute("func").pyobject
    assert isinstance(obj, pyobjectsdef.PyFunction)
    assert repr(obj).startswith(
        '<rope.base.pyobjectsdef.PyFunction "pkg1.mod1::func" at 0x'
    )


def test_repr_pyobjectsdef_pyfunction_without_associated_resource(project):
    code = """def func(arg): pass"""
    mod = libutils.get_string_module(project, code)
    obj = mod.get_attribute("func").pyobject
    assert isinstance(obj, pyobjectsdef.PyFunction)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyFunction "::func" at 0x')


def test_repr_pyobjectsdef_pyclass(project, mod1):
    code = """class MyClass: pass"""
    mod = libutils.get_string_module(project, code, mod1)
    obj = mod.get_attribute("MyClass").pyobject
    assert isinstance(obj, pyobjectsdef.PyClass)
    assert repr(obj).startswith(
        '<rope.base.pyobjectsdef.PyClass "pkg1.mod1::MyClass" at 0x'
    )


def test_repr_pyobjectsdef_pyclass_without_associated_resource(project):
    code = """class MyClass: pass"""
    mod = libutils.get_string_module(project, code)
    obj = mod.get_attribute("MyClass").pyobject
    assert isinstance(obj, pyobjectsdef.PyClass)
    assert repr(obj).startswith('<rope.base.pyobjectsdef.PyClass "::MyClass" at 0x')


def test_repr_pyobjectsdef_pycomprehension(project, mod1):
    code = """[a for a in b]"""
    mod = libutils.get_string_module(project, code, mod1)
    mod._create_structural_attributes()
    assert len(mod.defineds) == 1
    obj = mod.defineds[0]
    assert isinstance(obj, pyobjectsdef.PyComprehension)
    assert repr(obj).startswith(
        '<rope.base.pyobjectsdef.PyComprehension "pkg1.mod1::<comprehension>" at 0x'
    )


def test_repr_pyobjectsdef_pycomprehension_without_associated_resource(project):
    code = """[a for a in b]"""
    mod = libutils.get_string_module(project, code)
    mod._create_structural_attributes()
    assert len(mod.defineds) == 1
    obj = mod.defineds[0]
    assert isinstance(obj, pyobjectsdef.PyComprehension)
    assert repr(obj).startswith(
        '<rope.base.pyobjectsdef.PyComprehension "::<comprehension>" at 0x'
    )



import ast
from ast import *  # noqa: F401,F403

from rope.base import fscommands


def parse(source, filename="<string>", *args, **kwargs):  # type: ignore
    if isinstance(source, str):
        source = fscommands.unicode_to_file_data(source)
    if b"\r" in source:
        source = source.replace(b"\r\n", b"\n").replace(b"\r", b"\n")
    if not source.endswith(b"\n"):
        source += b"\n"
    try:
        return ast.parse(source, filename=filename, *args, **kwargs)
    except (TypeError, ValueError) as e:
        error = SyntaxError()
        error.lineno = 1
        error.filename = filename
        error.msg = str(e)
        raise error


def call_for_nodes(node, callback, recursive=False):
    """If callback returns `True` the child nodes are skipped"""
    result = callback(node)
    if recursive and not result:
        for child in ast.iter_child_nodes(node):
            call_for_nodes(child, callback, recursive)


class RopeNodeVisitor(ast.NodeVisitor):
    def visit(self, node):
        """Modified from ast.NodeVisitor to match rope's existing Visitor implementation"""
        method = "_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)



# The default ``config.py``
# flake8: noqa


def set_prefs(prefs):
    """This function is called before opening the project"""

    # Specify which files and folders to ignore in the project.
    # Changes to ignored resources are not added to the history and
    # VCSs.  Also they are not returned in `Project.get_files()`.
    # Note that ``?`` and ``*`` match all characters but slashes.
    # "*.pyc": matches "test.pyc" and "pkg/test.pyc"
    # "mod*.pyc": matches "test/mod1.pyc" but not "mod/1.pyc"
    # ".svn": matches "pkg/.svn" and all of its children
    # "build/*.o": matches "build/lib.o" but not "build/sub/lib.o"
    # "build//*.o": matches "build/lib.o" and "build/sub/lib.o"
    #
    #    prefs["ignored_resources"] = [
    #        "*.pyc",
    #        "*~",
    #        ".ropeproject",
    #        ".hg",
    #        ".svn",
    #        "_svn",
    #        ".git",
    #        ".tox",
    #        ".venv",
    #        "venv",
    #        ".mypy_cache",
    #        ".pytest_cache",
    #    ]

    # Specifies which files should be considered python files.  It is
    # useful when you have scripts inside your project.  Only files
    # ending with ``.py`` are considered to be python files by
    # default.
    #
    #     prefs["python_files"] = ["*.py"]

    # Custom source folders:  By default rope searches the project
    # for finding source folders (folders that should be searched
    # for finding modules).  You can add paths to that list.  Note
    # that rope guesses project source folders correctly most of the
    # time; use this if you have any problems.
    # The folders should be relative to project root and use "/" for
    # separating folders regardless of the platform rope is running on.
    # "src/my_source_folder" for instance.
    #
    #     prefs.add("source_folders", "src")

    # You can extend python path for looking up modules
    #
    #     prefs.add("python_path", "~/python/")

    # Should rope save object information or not.
    #
    #     prefs["save_objectdb"] = True
    #     prefs["compress_objectdb"] = False

    # If `True`, rope analyzes each module when it is being saved.
    #
    #     prefs["automatic_soa"] = True
    #
    # The depth of calls to follow in static object analysis
    #
    #     prefs["soa_followed_calls"] = 0

    # If `False` when running modules or unit tests "dynamic object
    # analysis" is turned off.  This makes them much faster.
    #
    #     prefs["perform_doa"] = True

    # Rope can check the validity of its object DB when running.
    #
    #    prefs["validate_objectdb"] = True

    # How many undos to hold?
    #
    #     prefs["max_history_items"] = 32

    # Shows whether to save history across sessions.
    #
    #     prefs["save_history"] = True
    #     prefs["compress_history"] = False

    # Set the number spaces used for indenting.  According to
    # :PEP:`8`, it is best to use 4 spaces.  Since most of rope's
    # unit-tests use 4 spaces it is more reliable, too.
    #
    #     prefs["indent_size"] = 4

    # Builtin and c-extension modules that are allowed to be imported
    # and inspected by rope.
    #
    #     prefs["extension_modules"] = []

    # Add all standard c-extensions to extension_modules list.
    #
    #     prefs["import_dynload_stdmods"] = True

    # If `True` modules with syntax errors are considered to be empty.
    # The default value is `False`; When `False` syntax errors raise
    # `rope.base.exceptions.ModuleSyntaxError` exception.
    #
    #     prefs["ignore_syntax_errors"] = False

    # If `True`, rope ignores unresolvable imports.  Otherwise, they
    # appear in the importing namespace.
    #
    #     prefs["ignore_bad_imports"] = False

    # If `True`, rope will insert new module imports as
    # `from <package> import <module>` by default.
    #
    #     prefs["prefer_module_from_imports"] = False

    # If `True`, rope will transform a comma list of imports into
    # multiple separate import statements when organizing
    # imports.
    #
    #     prefs["split_imports"] = False

    # If `True`, rope will remove all top-level import statements and
    # reinsert them at the top of the module when making changes.
    #
    #     prefs["pull_imports_to_top"] = True

    # If `True`, rope will sort imports alphabetically by module name instead
    # of alphabetically by import statement, with from imports after normal
    # imports.
    #
    #     prefs["sort_imports_alphabetically"] = False

    # Location of implementation of
    # rope.base.oi.type_hinting.interfaces.ITypeHintingFactory In general
    # case, you don't have to change this value, unless you're an rope expert.
    # Change this value to inject you own implementations of interfaces
    # listed in module rope.base.oi.type_hinting.providers.interfaces
    # For example, you can add you own providers for Django Models, or disable
    # the search type-hinting in a class hierarchy, etc.
    #
    #     prefs[
    #         "type_hinting_factory"
    #     ] = "rope.base.oi.type_hinting.factory.default_type_hinting_factory"


def project_opened(project):
    """This function is called after opening the project"""
    # Do whatever you like here!



import bisect
import difflib
import warnings

import rope.base.libutils
import rope.base.resourceobserver
import rope.base.resources
import rope.base.oi.doa
import rope.base.oi.objectinfo
import rope.base.oi.soa
from rope.base import (
    builtins,
    exceptions,
    pyobjectsdef,
    stdmods,
    taskhandle,
    utils,
)


class PyCore:
    def __init__(self, project):
        self.project = project
        self._init_resource_observer()
        self.cache_observers = []
        self.module_cache = _ModuleCache(self)
        self.extension_cache = _ExtensionCache(self)
        self.object_info = rope.base.oi.objectinfo.ObjectInfoManager(project)
        self._init_python_files()
        self._init_automatic_soa()

    def _init_python_files(self):
        self.python_matcher = None
        patterns = self.project.prefs.get("python_files", None)
        if patterns is not None:
            self.python_matcher = rope.base.resources._ResourceMatcher()
            self.python_matcher.set_patterns(patterns)

    def _init_resource_observer(self):
        callback = self._invalidate_resource_cache
        observer = rope.base.resourceobserver.ResourceObserver(
            changed=callback, moved=callback, removed=callback
        )
        self.observer = rope.base.resourceobserver.FilteredResourceObserver(observer)
        self.project.add_observer(self.observer)

    def _init_automatic_soa(self):
        if not self.automatic_soa:
            return
        callback = self._file_changed_for_soa
        observer = rope.base.resourceobserver.ResourceObserver(
            changed=callback, moved=callback, removed=callback
        )
        self.project.add_observer(observer)

    @property
    def automatic_soa(self):
        auto_soa = self.project.prefs.get("automatic_soi", None)
        return self.project.prefs.get("automatic_soa", auto_soa)

    def _file_changed_for_soa(self, resource, new_resource=None):
        old_contents = self.project.history.contents_before_current_change(resource)
        if old_contents is not None:
            perform_soa_on_changed_scopes(self.project, resource, old_contents)

    def is_python_file(self, resource):
        if resource.is_folder():
            return False
        if self.python_matcher is None:
            return resource.name.endswith(".py")
        return self.python_matcher.does_match(resource)

    @utils.deprecated("Use `project.get_module` instead")
    def get_module(self, name, folder=None):
        """Returns a `PyObject` if the module was found."""
        return self.project.get_module(name, folder)

    def _builtin_submodules(self, modname):
        result = {}
        for extension in self.extension_modules:
            if extension.startswith(modname + "."):
                name = extension[len(modname) + 1 :]
                if "." not in name:
                    result[name] = self.builtin_module(extension)
        return result

    def builtin_module(self, name):
        return self.extension_cache.get_pymodule(name)

    @utils.deprecated("Use `project.get_relative_module` instead")
    def get_relative_module(self, name, folder, level):
        return self.project.get_relative_module(name, folder, level)

    @utils.deprecated("Use `libutils.get_string_module` instead")
    def get_string_module(self, code, resource=None, force_errors=False):
        """Returns a `PyObject` object for the given code

        If `force_errors` is `True`, `exceptions.ModuleSyntaxError` is
        raised if module has syntax errors.  This overrides
        ``ignore_syntax_errors`` project config.

        """
        return pyobjectsdef.PyModule(self, code, resource, force_errors=force_errors)

    @utils.deprecated("Use `libutils.get_string_scope` instead")
    def get_string_scope(self, code, resource=None):
        """Returns a `Scope` object for the given code"""
        return rope.base.libutils.get_string_scope(code, resource)

    def _invalidate_resource_cache(self, resource, new_resource=None):
        for observer in self.cache_observers:
            observer(resource)

    @utils.deprecated("Use `project.get_python_path_folders` instead")
    def get_python_path_folders(self):
        return self.project.get_python_path_folders()

    @utils.deprecated("Use `project.find_module` instead")
    def find_module(self, modname, folder=None):
        """Returns a resource corresponding to the given module

        returns None if it can not be found
        """
        return self.project.find_module(modname, folder)

    @utils.deprecated("Use `project.find_relative_module` instead")
    def find_relative_module(self, modname, folder, level):
        return self.project.find_relative_module(modname, folder, level)

    # INFO: It was decided not to cache source folders, since:
    #  - Does not take much time when the root folder contains
    #    packages, that is most of the time
    #  - We need a separate resource observer; `self.observer`
    #    does not get notified about module and folder creations
    @utils.deprecated("Use `project.get_source_folders` instead")
    def get_source_folders(self):
        """Returns project source folders"""
        return self.project.get_source_folders()

    def resource_to_pyobject(self, resource, force_errors=False):
        return self.module_cache.get_pymodule(resource, force_errors)

    @utils.deprecated("Use `project.get_python_files` instead")
    def get_python_files(self):
        """Returns all python files available in the project"""
        return self.project.get_python_files()

    def _is_package(self, folder):
        return (
            folder.has_child("__init__.py")
            and not folder.get_child("__init__.py").is_folder()
        )

    def _find_source_folders(self, folder):
        for resource in folder.get_folders():
            if self._is_package(resource):
                return [folder]
        result = []
        for resource in folder.get_files():
            if resource.name.endswith(".py"):
                result.append(folder)
                break
        for resource in folder.get_folders():
            result.extend(self._find_source_folders(resource))
        return result

    def run_module(self, resource, args=None, stdin=None, stdout=None):
        """Run `resource` module

        Returns a `rope.base.oi.doa.PythonFileRunner` object for
        controlling the process.

        """
        perform_doa = self.project.prefs.get("perform_doi", True)
        perform_doa = self.project.prefs.get("perform_doa", perform_doa)
        receiver = self.object_info.doa_data_received
        if not perform_doa:
            receiver = None
        runner = rope.base.oi.doa.PythonFileRunner(
            self, resource, args, stdin, stdout, receiver
        )
        runner.add_finishing_observer(self.module_cache.forget_all_data)
        runner.run()
        return runner

    def analyze_module(
        self,
        resource,
        should_analyze=lambda py: True,
        search_subscopes=lambda py: True,
        followed_calls=None,
    ):
        """Analyze `resource` module for static object inference

        This function forces rope to analyze this module to collect
        information about function calls.  `should_analyze` is a
        function that is called with a `PyDefinedObject` argument.  If
        it returns `True` the element is analyzed.  If it is `None` or
        returns `False` the element is not analyzed.

        `search_subscopes` is like `should_analyze`; The difference is
        that if it returns `False` the sub-scopes are all ignored.
        That is it is assumed that `should_analyze` returns `False`
        for all of its subscopes.

        `followed_calls` override the value of ``soa_followed_calls``
        project config.
        """
        if followed_calls is None:
            followed_calls = self.project.prefs.get("soa_followed_calls", 0)
        pymodule = self.resource_to_pyobject(resource)
        self.module_cache.forget_all_data()
        rope.base.oi.soa.analyze_module(
            self, pymodule, should_analyze, search_subscopes, followed_calls
        )

    def get_classes(self, task_handle=taskhandle.NullTaskHandle()):
        warnings.warn(
            "`PyCore.get_classes()` is deprecated", DeprecationWarning, stacklevel=2
        )
        return []

    def __str__(self):
        return str(self.module_cache) + str(self.object_info)

    @utils.deprecated("Use `libutils.modname` instead")
    def modname(self, resource):
        return rope.base.libutils.modname(resource)

    @property
    @utils.cacheit
    def extension_modules(self):
        result = set(self.project.prefs.get("extension_modules", []))
        if self.project.prefs.get("import_dynload_stdmods", False):
            result.update(stdmods.dynload_modules())
        return result


class _ModuleCache:
    def __init__(self, pycore):
        self.pycore = pycore
        self.module_map = {}
        self.pycore.cache_observers.append(self._invalidate_resource)
        self.observer = self.pycore.observer

    def _invalidate_resource(self, resource):
        if resource in self.module_map:
            self.forget_all_data()
            self.observer.remove_resource(resource)
            del self.module_map[resource]

    def get_pymodule(self, resource, force_errors=False):
        if resource in self.module_map:
            return self.module_map[resource]
        if resource.is_folder():
            result = pyobjectsdef.PyPackage(
                self.pycore,
                resource,
                force_errors=force_errors,
            )
        else:
            result = pyobjectsdef.PyModule(
                self.pycore,
                resource=resource,
                force_errors=force_errors,
            )
            if result.has_errors:
                return result
        self.module_map[resource] = result
        self.observer.add_resource(resource)
        return result

    def forget_all_data(self):
        for pymodule in self.module_map.values():
            pymodule._forget_concluded_data()

    def __str__(self):
        return "PyCore caches %d PyModules\n" % len(self.module_map)


class _ExtensionCache:
    def __init__(self, pycore):
        self.pycore = pycore
        self.extensions = {}

    def get_pymodule(self, name):
        if name == "__builtin__":
            return builtins.builtins
        allowed = self.pycore.extension_modules
        if name not in self.extensions and name in allowed:
            self.extensions[name] = builtins.BuiltinModule(name, self.pycore)
        return self.extensions.get(name)


def perform_soa_on_changed_scopes(project, resource, old_contents):
    pycore = project.pycore
    if resource.exists() and pycore.is_python_file(resource):
        try:
            new_contents = resource.read()
            # detecting changes in new_contents relative to old_contents
            detector = _TextChangeDetector(new_contents, old_contents)

            def search_subscopes(pydefined):
                scope = pydefined.get_scope()
                return detector.is_changed(scope.get_start(), scope.get_end())

            def should_analyze(pydefined):
                scope = pydefined.get_scope()
                start = scope.get_start()
                end = scope.get_end()
                return detector.consume_changes(start, end)

            pycore.analyze_module(resource, should_analyze, search_subscopes)
        except exceptions.ModuleSyntaxError:
            pass


class _TextChangeDetector:
    def __init__(self, old, new):
        self.old = old
        self.new = new
        self._set_diffs()

    def _set_diffs(self):
        differ = difflib.Differ()
        self.lines = []
        lineno = 0
        for line in differ.compare(
            self.old.splitlines(True), self.new.splitlines(True)
        ):
            if line.startswith(" "):
                lineno += 1
            elif line.startswith("-"):
                lineno += 1
                self.lines.append(lineno)

    def is_changed(self, start, end):
        """Tell whether any of start till end lines have changed

        The end points are inclusive and indices start from 1.
        """
        left, right = self._get_changed(start, end)
        return left < right

    def consume_changes(self, start, end):
        """Clear the changed status of lines from start till end"""
        left, right = self._get_changed(start, end)
        if left < right:
            del self.lines[left:right]
        return left < right

    def _get_changed(self, start, end):
        left = bisect.bisect_left(self.lines, start)
        right = bisect.bisect_right(self.lines, end)
        return left, right



from typing import List


class FinalQuery:
    def __init__(self, query):
        self._query = query


class Query:
    def __init__(self, query: str, columns: List[str]):
        self.query = query
        self.columns = columns

    def select(self, *columns: str):
        if not (set(columns) <= set(self.columns)):
            raise ValueError(
                f"Unknown column names passed: {set(columns) - set(self.columns)}"
            )

        selected_columns = ", ".join(columns)
        return FinalQuery(f"SELECT {selected_columns} FROM {self.query}")

    def select_star(self):
        return FinalQuery(f"SELECT * FROM {self.query}")

    def where(self, where_clause: str):
        return Query(
            f"{self.query} WHERE {where_clause}",
            columns=self.columns,
        )

    def insert_into(self) -> FinalQuery:
        columns = ", ".join(self.columns)
        placeholders = ", ".join(["?"] * len(self.columns))
        return FinalQuery(
            f"INSERT INTO {self.query}({columns}) VALUES ({placeholders})"
        )

    def drop_table(self) -> FinalQuery:
        return FinalQuery(f"DROP TABLE {self.query}")

    def delete_from(self) -> FinalQuery:
        return FinalQuery(f"DELETE FROM {self.query}")


class Name:
    table_name = "names"
    columns = [
        "name",
        "module",
        "package",
        "source",
        "type",
    ]

    @classmethod
    def create_table(self, connection):
        names_table = (
            "(name TEXT, module TEXT, package TEXT, source INTEGER, type INTEGER)"
        )
        connection.execute(f"CREATE TABLE IF NOT EXISTS names{names_table}")
        connection.execute("CREATE INDEX IF NOT EXISTS name ON names(name)")
        connection.execute("CREATE INDEX IF NOT EXISTS module ON names(module)")
        connection.execute("CREATE INDEX IF NOT EXISTS package ON names(package)")

    objects = Query(table_name, columns)

    search_submodule_like = objects.where('module LIKE ("%." || ?)')
    search_module_like = objects.where("module LIKE (?)")

    import_assist = objects.where("name LIKE (? || '%')")

    search_by_name_like = objects.where("name LIKE (?)")

    delete_by_module_name = objects.where("module = ?").delete_from()


class Package:
    table_name = "packages"
    columns = [
        "package",
        "path",
    ]

    @classmethod
    def create_table(self, connection):
        packages_table = "(package TEXT, path TEXT)"
        connection.execute(f"CREATE TABLE IF NOT EXISTS packages{packages_table}")

    objects = Query(table_name, columns)

    delete_by_package_name = objects.where("package = ?").delete_from()
